-- store the last feed id
create temp table _last_feed as
select
  feed_id
  , id
from
  tempus_gtfs.feed_info
where
  id = (select max(id) from tempus_gtfs.feed_info);

--
-- compute stop geometries
do $$
begin
raise notice '==== compute stop geometries ===';
end$$;

update
  tempus_gtfs.stops
set
  geom = st_setsrid(st_makepoint(stop_lon, stop_lat, 0), 4326)
;

do $$
begin
raise notice '==== road section spatial matching ===';
end$$;

--
-- 1. attach each stop to the closest road section in a 50 meters radius
--

-- first, create an index on road_section geography 
create index on tempus.road_section using gist(geography(geom));

-- actual update
update
  tempus_gtfs.stops
set
  road_section_id = t.rs_id
  , abscissa_road_section = t.abscissa
from
(
select
  -- distinct will select the first row from the order by
  distinct on (stop_id)
  stops.stop_id
  , rs.id rs_id
  , st_distance(stops.geom,rs.geom) dist
  , st_linelocatepoint(rs.geom, stops.geom) as abscissa
from
  tempus_gtfs.stops
  -- convert to geography to compare distances in meters
  join tempus.road_section rs on st_dwithin(geography(stops.geom), geography(rs.geom), 50)
where
  -- attach to roads waklable by pedestrians
  (rs.traffic_rules_ft & 1) > 0 or (rs.traffic_rules_tf & 1) > 0
order by
  stop_id, dist
) t
where
  t.stop_id = stops.stop_id
  -- only for the latest feed
  and stops.feed_id = (select feed_id from _last_feed limit 1)
;

drop index tempus.road_section_geography_idx;

do $$
begin
raise notice '==== new road sections for unmatched stops ===';
end$$;

drop sequence if exists tempus.seq_road_node_id;
create sequence tempus.seq_road_node_id start with 1;
select setval('tempus.seq_road_node_id', (select max(id) from tempus.road_node));

drop sequence if exists tempus.seq_road_section_id;
create sequence tempus.seq_road_section_id start with 1;
select setval('tempus.seq_road_section_id', (select max(id) from tempus.road_section));

--
-- the new roads to create
drop table if exists _new_roads;
create table _new_roads as
select
  stop_id
  , nextval('tempus.seq_road_node_id')::bigint as node1_id
  , nextval('tempus.seq_road_node_id')::bigint as node2_id
  , nextval('tempus.seq_road_section_id')::bigint as rs_id
  , st_translate(stops.geom, -0.0001, 0, 0) as node1_geom
  , st_translate(stops.geom, 0.0001, 0, 0) as node2_geom
from
  tempus_gtfs.stops
where
  road_section_id is null
;

-- new road nodes
insert into tempus.road_node
select
  node1_id
  , false as bifurcation
  , node1_geom
from
  _new_roads
union all
select
  node2_id
  , false as bifurcation
  , node2_geom
from
  _new_roads
;

-- new road sections
insert into tempus.road_section
  (id, road_type, node_from, node_to, traffic_rules_ft, traffic_rules_tf, length, car_speed_limit, road_name, lane, roundabout, bridge, tunnel, ramp, tollway, geom)
select
  rs_id
  , 1 as road_type
  , node1_id as node_from
  , node2_id as node_to
  , 32767 as traffic_rules_ft
  , 32767 as traffic_rules_tf
  , 0 as length
  , 0 as car_speed_limit
  , '' as road_name
  , 1 as lane
  , false as roundabout
  , false as bridge
  , false as tunnel
  , false as ramp
  , false as tollway
  , st_makeline(node1_geom, node2_geom)
from
  _new_roads
;

-- attach stops to new road sections
update
  tempus_gtfs.stops
set
  road_section_id = rs_id
  , abscissa_road_section = 0.5
  , artificial_road_section = true
from
  _new_roads
where
  stops.stop_id = _new_roads.stop_id
;

--drop table _tempus_import.new_roads;


do $$
begin
raise notice '==== add new transport modes ===';
end$$;

--
-- Create new transport modes
--
drop sequence if exists transport_mode_id_seq;
create sequence transport_mode_id_seq start with 1;
select setval('transport_mode_id_seq', (select max(id) from tempus.transport_mode));

insert into tempus.transport_mode(id, name, public_transport, gtfs_route_type, gtfs_feed_id)
select
        nextval('transport_mode_id_seq') as id
        , (case
        when r.route_type = 0 then 'Tram'
        when r.route_type = 1 then 'Subway'
        when r.route_type = 2 then 'Train'
        when r.route_type = 3 then 'Bus'
        when r.route_type = 4 then 'Ferry'
        when r.route_type = 5 then 'Cable-car'
        when r.route_type = 6 then 'Suspended Cable-Car'
        when r.route_type = 7 then 'Funicular'
        end) || ' (' || (select feed_id from _last_feed limit 1) || ')'
	, true
	, r.route_type
        , (select id from _last_feed limit 1)
from (
  select distinct route_type from tempus_gtfs.routes
  -- only routes from last feed
  where feed_id = (select feed_id from _last_feed limit 1)
) r
;


do $$
begin
raise notice '==== compute shapes ===';
end$$;

update
  tempus_gtfs.shapes
set
  geom = t.geom
from
(
  select
    shape_id
    , st_makeline(array_agg(st_setsrid(st_makepoint(shape_pt_lon, shape_pt_lat, 0.0), 4326) order by shape_pt_sequence)) as geom
  from
    tempus_gtfs.shape_pts
  group by
    shape_id
) t
where
  shapes.shape_id = t.shape_id
  and shapes.feed_id = (select feed_id from _last_feed limit 1)
;

do $$
begin
raise notice '==== compute sections ===';
end$$;

insert into tempus_gtfs.sections (feed_id, stop_from, stop_to, geom)
select
  (select id from _last_feed limit 1)
  , st1.id
  , st2.id
  , st_makeline(st1.geom, st2.geom)
from
(  
select
  distinct on (st1.stop_id, st2.stop_id)
  st1.stop_id as stop1
  , st2.stop_id as stop2
from
  tempus_gtfs.stop_times st1
  join tempus_gtfs.stop_times st2 on (st1.trip_id = st2.trip_id) and (st1.feed_id = st2.feed_id) and (st2.stop_sequence = st1.stop_sequence + 1)
where
  st1.feed_id = (select feed_id from _last_feed limit 1)
) t
join tempus_gtfs.stops st1 on st1.stop_id = stop1
join tempus_gtfs.stops st2 on st2.stop_id = stop2
;

