-- Tempus Database schema: version 1.1
--

--
-- DROP and clean if needed
--
CREATE EXTENSION IF NOT EXISTS postgis;

DROP SCHEMA IF EXISTS tempus CASCADE;
DELETE FROM public.geometry_columns WHERE f_table_schema='tempus';
DROP SCHEMA IF EXISTS tempus_gtfs CASCADE;


--
-- Tempus Schema
--
CREATE SCHEMA tempus;

CREATE TABLE tempus.transport_mode
(
    id serial PRIMARY KEY,
    name varchar, -- Description of the mode
    public_transport boolean NOT NULL,
    gtfs_route_type integer, -- Reference to the equivalent GTFS codification (for PT only)
    traffic_rules integer, -- Binary composition of TransportModeTrafficRule
    speed_rule integer, -- TransportModeSpeedRule
    toll_rule integer, -- Binary composition of TransportModeToolRule
    engine_type integer, -- TransportModeEngine
    need_parking boolean,
    shared_vehicle boolean,
    return_shared_vehicle boolean
);

COMMENT ON TABLE tempus.transport_mode IS 'Available transport modes';
COMMENT ON COLUMN tempus.transport_mode.name IS 'Description of the mode';
COMMENT ON COLUMN tempus.transport_mode.traffic_rules IS 'Bitfield value: defines road traffic rules followed by the mode, NULL for PT modes';
COMMENT ON COLUMN tempus.transport_mode.gtfs_route_type IS 'Reference to the equivalent GTFS code (for PT only)';
COMMENT ON COLUMN tempus.transport_mode.speed_rule IS 'Defines the road speed rule followed by the mode, NULL for PT modes';
COMMENT ON COLUMN tempus.transport_mode.toll_rule IS 'Bitfield value: gives the toll rules followed by the mode, NULL for PT modes';
COMMENT ON COLUMN tempus.transport_mode.need_parking IS 'If vehicle needs to be parked, NULL for PT modes';
COMMENT ON COLUMN tempus.transport_mode.shared_vehicle IS 'If vehicule is shared and needs to be return at a/some stations at the end of the trip, NULL for PT modes';
COMMENT ON COLUMN tempus.transport_mode.return_shared_vehicle IS 'If vehicule is shared and needs to be returned to its initial station at the end of a loop, NULL for PT modes';
-- TODO Add a CHECK on parent_id related to id bitfield values

INSERT INTO tempus.transport_mode(name, public_transport, gtfs_route_type, traffic_rules, speed_rule, toll_rule, engine_type, need_parking, shared_vehicle, return_shared_vehicle)
	VALUES ('Walking',         'f', NULL, 1,  1, NULL, NULL, 'f', 'f', 'f');
INSERT INTO tempus.transport_mode(name, public_transport, gtfs_route_type, traffic_rules, speed_rule, toll_rule, engine_type, need_parking, shared_vehicle, return_shared_vehicle)
	VALUES ('Private bicycle', 'f', NULL, 2,  2, NULL, NULL, 't', 'f', 'f');
INSERT INTO tempus.transport_mode(name, public_transport, gtfs_route_type, traffic_rules, speed_rule, toll_rule, engine_type, need_parking, shared_vehicle, return_shared_vehicle)
	VALUES ('Private car',     'f', NULL, 4,  5, 1,    1,    't', 'f', 'f');
INSERT INTO tempus.transport_mode(name, public_transport, gtfs_route_type, traffic_rules, speed_rule, toll_rule, engine_type, need_parking, shared_vehicle, return_shared_vehicle)
        VALUES ('Taxi',            'f', NULL, 8, 5, 1,    1,    'f', 'f', 'f');

CREATE TABLE tempus.road_validity_period
(
    id integer PRIMARY KEY,
    name varchar,
    monday boolean DEFAULT true,
    tuesday boolean DEFAULT true,
    wednesday boolean DEFAULT true,
    thursday boolean DEFAULT true,
    friday boolean DEFAULT true,
    saturday boolean DEFAULT true,
    sunday boolean DEFAULT true,
    bank_holiday boolean DEFAULT true,
    day_before_bank_holiday boolean DEFAULT true,
    holidays boolean DEFAULT true,
    day_before_holidays boolean DEFAULT true,
    start_date date,
    end_date date
);
COMMENT ON TABLE tempus.road_validity_period IS 'Periods during which road restrictions and speed profiles apply';
INSERT INTO tempus.road_validity_period VALUES (0, 'Always', true, true, true, true, true, true, true, true, true, true, true, NULL, NULL);

/*
CREATE TABLE tempus.seasonal_ticket
(
    id integer PRIMARY KEY, --bitfield
    name varchar NOT NULL,
    price double precision,
    people_concerned varchar
);
INSERT INTO tempus.seasonal_ticket VALUES (1, 'Shared bicycle day ticket', 1.5, 'everybody');
INSERT INTO tempus.seasonal_ticket VALUES (2, 'Shared bicycle week ticket', 5, 'everybody');
INSERT INTO tempus.seasonal_ticket VALUES (4, 'Shared bicycle year ticket', 25, 'everybody');



CREATE TABLE tempus.road_vehicle_fare_rule
(
    id integer PRIMARY KEY,
    name varchar NOT NULL,
    seasonal_ticket integer,  -- bitfield
    transport_type integer,
    price_per_km double precision,
    price_per_minute double precision,
    price_per_use double precision,
    min_minutes integer,
    max_mintes integer
    min_km integer,
    max_km integer,
    start_time time without time zone,
    end_time time without time zone
);

INSERT INTO tempus.road_vehicle_fare_rule VALUES (1, 'Shared bicycle first half an hour of use', 7, 0, 0, 0, 0, 30, NULL, NULL, NULL, NULL);
INSERT INTO tempus.road_vehicle_fare_rule VALUES (2, 'Shared bicycle second half an hour of use', 7, 0, 0, 1, 30, 60, NULL, NULL, NULL, NULL);
INSERT INTO tempus.road_vehicle_fare_rule VALUES (3, 'Shared bicycle third half an hour of use', 7, 0, 0, 2, 60, 90, NULL, NULL, NULL, NULL);
INSERT INTO tempus.road_vehicle_fare_rule VALUES (4, 'Shared bicycle fourth half an hour of use', 7, 0, 0, 2, 90, 120, NULL, NULL, NULL, NULL);


*/
--TODO: add a data model able to represent taxis and shared vehicles fare rules => be able to give marginal cost of transports for any user category (with or without subscription to transport services...)


CREATE TABLE tempus.bank_holiday
(
    calendar_date date PRIMARY KEY,
    name varchar
);
COMMENT ON TABLE tempus.bank_holiday IS 'Bank holiday list';


CREATE TABLE tempus.holidays
(
    id serial,
    name varchar,
    start_date date,
    end_date date
);
COMMENT ON TABLE tempus.holidays IS 'Holidays description';


--
-- Roads
-- NOTA: Consider look at OSM specification:
--       <http://wiki.openstreetmap.org/wiki/Map_Features>
--	 <http://wiki.openstreetmap.org/wiki/Tagging_samples/urban>
--       and
--       "Description des bases de donnees vectorielles routieres" DREAL NPdC

CREATE TABLE tempus.road_node
(
    id bigint PRIMARY KEY,
    bifurcation boolean -- total number of incident edges is > 2
    -- NOTA: geometry column added NOT NULL
);
SELECT AddGeometryColumn('tempus', 'road_node', 'geom', 4326, 'POINT', 3);
COMMENT ON TABLE tempus.road_node IS 'Road nodes description';
COMMENT ON COLUMN tempus.road_node.bifurcation IS 'If true, total number of incident edges is > 2';


CREATE TABLE tempus.road_section
(
    id bigint PRIMARY KEY,
    vendor_id bigint,
    road_type integer,
    node_from bigint NOT NULL REFERENCES tempus.road_node ON DELETE CASCADE ON UPDATE CASCADE DEFERRABLE INITIALLY IMMEDIATE,
    node_to bigint NOT NULL REFERENCES tempus.road_node ON DELETE CASCADE ON UPDATE CASCADE DEFERRABLE INITIALLY IMMEDIATE,
    traffic_rules_ft smallint NOT NULL, -- References tempus.road_traffic_rule => bitfield value
    traffic_rules_tf smallint NOT NULL, -- References tempus.road_traffic_rule => bitfield value
    length double precision NOT NULL, -- in meters
    car_speed_limit double precision, -- in km/h
    road_name varchar,
    lane integer,
    roundabout boolean,
    bridge boolean,
    tunnel boolean,
    ramp boolean, -- or sliproads
    tollway boolean
    -- NOTA: geometry column added NOT NULL
);
SELECT AddGeometryColumn('tempus', 'road_section', 'geom', 4326, 'LINESTRING', 3);
COMMENT ON TABLE tempus.road_section IS 'Road sections description';
-- TODO Add a CHECK on transport_type_* bitfields value
COMMENT ON COLUMN tempus.road_section.road_type IS '1: fast links between urban areas, 2: links between 1 level links, heavy traffic with lower speeds, 3: local links with heavy traffic, 4: low traffic';
COMMENT ON COLUMN tempus.road_section.traffic_rules_ft IS 'Bitfield value giving allowed traffic rules for direction from -> to';
COMMENT ON COLUMN tempus.road_section.traffic_rules_tf IS 'Bitfield value giving allowed traffic rules for direction to -> from';
COMMENT ON COLUMN tempus.road_section.length IS 'In meters';
COMMENT ON COLUMN tempus.road_section.car_speed_limit IS 'In km/h';
COMMENT ON COLUMN tempus.road_section.ramp IS 'Or sliproad';

CREATE TABLE tempus.road_daily_profile
(
    profile_id integer NOT NULL,
    begin_time integer NOT NULL,
    speed_rule integer NOT NULL,
    end_time  integer NOt NULL,
    average_speed double precision NOT NULL, -- In km/h
    PRIMARY KEY (profile_id, speed_rule, begin_time)
);
COMMENT ON COLUMN tempus.road_daily_profile.begin_time IS 'When the period begins. Number of minutes since midnight';
COMMENT ON COLUMN tempus.road_daily_profile.end_time IS 'When the period ends. Number of minutes since midnight';
COMMENT ON COLUMN tempus.road_daily_profile.speed_rule IS 'Speed rule: car, truck, bike, etc.';
COMMENT ON COLUMN tempus.road_daily_profile.average_speed IS 'Speed value in km/h';

CREATE TABLE tempus.road_section_speed
(
    road_section_id bigint NOT NULL REFERENCES tempus.road_section ON DELETE CASCADE ON UPDATE CASCADE,
    period_id integer NOT NULL REFERENCES tempus.road_validity_period ON DELETE CASCADE ON UPDATE CASCADE,
    profile_id integer NOT NULL, -- road_daily_profile
    PRIMARY KEY (road_section_id, period_id, profile_id)
);
COMMENT ON TABLE tempus.road_section_speed IS 'Speed, vehicle types and validity period associated to road sections';
COMMENT ON COLUMN tempus.road_section_speed.period_id IS '0 if always applies';
COMMENT ON COLUMN tempus.road_section_speed.profile_id IS 'Reference to tempus.road_daily_profile';


CREATE TABLE tempus.road_restriction
(
    id bigint PRIMARY KEY,
    sections bigint[] NOT NULL
);
COMMENT ON TABLE tempus.road_restriction IS 'Road sections lists submitted to a restriction';
COMMENT ON COLUMN tempus.road_restriction.sections IS 'Involved road sections ID, not always forming a path';

CREATE TABLE tempus.road_restriction_time_penalty
(
    restriction_id bigint NOT NULL REFERENCES tempus.road_restriction ON DELETE CASCADE ON UPDATE CASCADE,
    period_id integer NOt NULL REFERENCES tempus.road_validity_period ON DELETE CASCADE ON UPDATE CASCADE, -- 0 if always applies
    traffic_rules integer NOT NULL, -- References tempus.road_traffic_rule => bitfield value
    time_value double precision NOT NULL,
    PRIMARY KEY (restriction_id, period_id, traffic_rules)
);
COMMENT ON TABLE tempus.road_restriction_time_penalty IS 'Time penalty (including infinite values for forbidden movements) applied to road restrictions';
COMMENT ON COLUMN tempus.road_restriction_time_penalty.period_id IS '0 if always applies';
COMMENT ON COLUMN tempus.road_restriction_time_penalty.traffic_rules IS 'References tempus.transport_mode_traffic_rule => Bitfield value';
COMMENT ON COLUMN tempus.road_restriction_time_penalty.time_value IS 'In minutes';


CREATE TABLE tempus.road_restriction_toll
(
    restriction_id bigint NOT NULL REFERENCES tempus.road_restriction ON DELETE CASCADE ON UPDATE CASCADE,
    period_id integer NOT NULL REFERENCES tempus.road_validity_period ON DELETE CASCADE ON UPDATE CASCADE, -- NULL if always applies
    toll_rules integer NOT NULL, -- References tempus.road_toll_rule => bitfield value
    toll_value double precision,
    PRIMARY KEY (restriction_id, period_id, toll_rules)
);
COMMENT ON TABLE tempus.road_restriction_toll IS 'Tolls applied to road restrictions';
COMMENT ON COLUMN tempus.road_restriction_toll.period_id IS '0 if always applies';
COMMENT ON COLUMN tempus.road_restriction_toll.toll_rules IS 'References tempus.transport_mode_toll_rule => Bitfield value, defines the type of vehicles to which it applies';
COMMENT ON COLUMN tempus.road_restriction_toll.toll_value IS 'In euros, can be NULL if unknown';

--
-- POI
--

CREATE TABLE tempus.poi
(
	id integer PRIMARY KEY,
	poi_type integer,
	name varchar,
        parking_transport_modes integer[] NOT NULL,
	road_section_id bigint REFERENCES tempus.road_section ON DELETE CASCADE ON UPDATE CASCADE,
	abscissa_road_section double precision NOT NULL CHECK (abscissa_road_section >= 0 AND abscissa_road_section <= 1)
	-- NOTA: geometry column added NOT NULL
);
SELECT AddGeometryColumn('tempus', 'poi', 'geom', 4326, 'POINT', 3);
COMMENT ON TABLE tempus.poi IS 'Points of Interest description';

--
-- GTFS' tables
--

---------- automatic extract from gtfslib ----------

CREATE SCHEMA tempus_gtfs;

CREATE TABLE tempus_gtfs.feed_info (
        feed_id VARCHAR NOT NULL,
        feed_publisher_name VARCHAR,
        feed_publisher_url VARCHAR,
        feed_contact_email VARCHAR,
        feed_contact_url VARCHAR,
        feed_lang VARCHAR,
        feed_start_date DATE,
        feed_end_date DATE,
        feed_version VARCHAR,
        PRIMARY KEY (feed_id)
);

CREATE TABLE tempus_gtfs.agency (
        feed_id VARCHAR NOT NULL,
        agency_id VARCHAR NOT NULL,
        agency_name VARCHAR NOT NULL,
        agency_url VARCHAR NOT NULL,
        agency_timezone VARCHAR NOT NULL,
        agency_lang VARCHAR,
        agency_phone VARCHAR,
        agency_fare_url VARCHAR,
        agency_email VARCHAR,
        PRIMARY KEY (feed_id, agency_id)
);

CREATE TABLE tempus_gtfs.calendar (
        feed_id VARCHAR NOT NULL,
        service_id VARCHAR NOT NULL,
        PRIMARY KEY (feed_id, service_id)
);

CREATE TABLE tempus_gtfs.fare_attributes (
        feed_id VARCHAR NOT NULL,
        fare_id VARCHAR NOT NULL,
        price FLOAT NOT NULL,
        currency_type VARCHAR NOT NULL,
        payment_method INTEGER NOT NULL,
        transfers INTEGER,
        transfer_duration INTEGER,
        PRIMARY KEY (feed_id, fare_id)
);

CREATE TABLE tempus_gtfs.shapes (
        feed_id VARCHAR NOT NULL,
        shape_id VARCHAR NOT NULL,
        PRIMARY KEY (feed_id, shape_id)
);

CREATE TABLE tempus_gtfs.zones (
        feed_id VARCHAR NOT NULL,
        zone_id VARCHAR NOT NULL,
        PRIMARY KEY (feed_id, zone_id)
);

CREATE TABLE tempus_gtfs.calendar_dates (
        feed_id VARCHAR NOT NULL,
        service_id VARCHAR NOT NULL,
        date DATE NOT NULL,
        PRIMARY KEY (feed_id, service_id, date)
);
CREATE INDEX idx_calendar_dates_date ON tempus_gtfs.calendar_dates(feed_id,date);

CREATE TABLE tempus_gtfs.fare_rules (
        feed_id VARCHAR NOT NULL,
        fare_id VARCHAR NOT NULL,
        route_id VARCHAR,
        origin_id VARCHAR,
        destination_id VARCHAR,
        contains_id VARCHAR,
        PRIMARY KEY (feed_id, fare_id, route_id, origin_id, destination_id, contains_id)
);

CREATE TABLE tempus_gtfs.routes (
        feed_id VARCHAR NOT NULL,
        route_id VARCHAR NOT NULL,
        agency_id VARCHAR NOT NULL,
        route_short_name VARCHAR,
        route_long_name VARCHAR,
        route_desc VARCHAR,
        route_type INTEGER NOT NULL,
        route_url VARCHAR,
        route_color VARCHAR,
        route_text_color VARCHAR,
        PRIMARY KEY (feed_id, route_id)
);
CREATE INDEX idx_routes_type ON tempus_gtfs.routes(feed_id,route_type);
CREATE INDEX idx_routes_agency ON tempus_gtfs.routes(feed_id,agency_id);
CREATE INDEX idx_routes_short_name ON tempus_gtfs.routes(feed_id,route_short_name);

CREATE TABLE tempus_gtfs.shape_pts (
        feed_id VARCHAR NOT NULL,
        shape_id VARCHAR NOT NULL,
        shape_pt_sequence INTEGER NOT NULL,
        shape_dist_traveled FLOAT NOT NULL,
        shape_pt_lat FLOAT NOT NULL,
        shape_pt_lon FLOAT NOT NULL,
        PRIMARY KEY (feed_id, shape_id, shape_pt_sequence)
);
CREATE INDEX idx_shape_pt_shape ON tempus_gtfs.shape_pts(feed_id,shape_id);

CREATE TABLE tempus_gtfs.stops (
        feed_id VARCHAR NOT NULL,
        stop_id VARCHAR NOT NULL,
        parent_station_id VARCHAR,
        location_type INTEGER NOT NULL,
        stop_name VARCHAR NOT NULL,
        stop_lat FLOAT NOT NULL,
        stop_lon FLOAT NOT NULL,
        wheelchair_boarding INTEGER NOT NULL,
        stop_code VARCHAR,
        stop_desc VARCHAR,
        zone_id VARCHAR,
        stop_url VARCHAR,
        stop_timezone VARCHAR,
        PRIMARY KEY (feed_id, stop_id)
);
CREATE INDEX idx_stops_zone ON tempus_gtfs.stops(feed_id,zone_id);
CREATE INDEX idx_stops_code ON tempus_gtfs.stops(feed_id,stop_code);
CREATE INDEX idx_stops_parent ON tempus_gtfs.stops(feed_id,parent_station_id);
CREATE INDEX idx_stops_lat ON tempus_gtfs.stops(stop_lat);
CREATE INDEX idx_stops_lon ON tempus_gtfs.stops(stop_lon);

CREATE TABLE tempus_gtfs.transfers (
        feed_id VARCHAR NOT NULL,
        from_stop_id VARCHAR NOT NULL,
        to_stop_id VARCHAR NOT NULL,
        transfer_type INTEGER NOT NULL,
        min_transfer_time INTEGER,
        PRIMARY KEY (feed_id, from_stop_id, to_stop_id)
);
CREATE INDEX idx_transfer_from ON tempus_gtfs.transfers(feed_id,from_stop_id);
CREATE INDEX idx_transfer_to ON tempus_gtfs.transfers(feed_id,to_stop_id);

CREATE TABLE tempus_gtfs.trips (
        feed_id VARCHAR NOT NULL,
        trip_id VARCHAR NOT NULL,
        route_id VARCHAR NOT NULL,
        service_id VARCHAR NOT NULL,
        shape_id VARCHAR,
        wheelchair_accessible INTEGER NOT NULL,
        bikes_allowed INTEGER NOT NULL,
        exact_times INTEGER NOT NULL,
        frequency_generated BOOLEAN NOT NULL,
        trip_headsign VARCHAR,
        trip_short_name VARCHAR,
        direction_id INTEGER,
        block_id VARCHAR,
        PRIMARY KEY (feed_id, trip_id)
);
CREATE INDEX idx_trips_route ON tempus_gtfs.trips(feed_id,route_id);
CREATE INDEX idx_trips_service ON tempus_gtfs.trips(feed_id,service_id);

CREATE TABLE tempus_gtfs.stop_times (
        feed_id VARCHAR NOT NULL,
        trip_id VARCHAR NOT NULL,
        stop_sequence INTEGER NOT NULL,
        stop_id VARCHAR NOT NULL,
        arrival_time INTEGER,
        departure_time INTEGER,
        interpolated BOOLEAN NOT NULL,
        shape_dist_traveled FLOAT NOT NULL,
        timepoint INTEGER NOT NULL,
        pickup_type INTEGER NOT NULL,
        drop_off_type INTEGER NOT NULL,
        stop_headsign VARCHAR,
        PRIMARY KEY (feed_id, trip_id, stop_sequence)
);
CREATE INDEX idx_stop_times_stop ON tempus_gtfs.stop_times(feed_id,stop_id);
CREATE INDEX idx_stop_times_sequence ON tempus_gtfs.stop_times(feed_id,stop_sequence);

---------- end of automatic extract from gtfslib ----------

--
-- Additional fields in GTFS' tables
--

-- a serial id to feed_info
alter table tempus_gtfs.feed_info
add column id serial unique not null;
create index on tempus_gtfs.feed_info(id);

-- a serial id to stops
alter table tempus_gtfs.stops
add column id serial unique not null;
create index on tempus_gtfs.stops(id);

-- a serial id to trips
alter table tempus_gtfs.trips
add column id serial unique not null;
create index on tempus_gtfs.trips(id);

-- stop geometry
select AddGeometryColumn('tempus_gtfs', 'stops', 'geom', 4326, 'POINT', 3);

-- stop -> road_section attachment
alter table tempus_gtfs.stops
add column road_section_id bigint references tempus.road_section on delete cascade on update cascade;
alter table tempus_gtfs.stops
add column abscissa_road_section double precision check (abscissa_road_section >= 0 and abscissa_road_section <= 1);
-- artificial road section: whether a road section has been created during import for this stop
alter table tempus_gtfs.stops
add column artificial_road_section boolean default false;

-- trigger to propagate stop deletion to artificial road sections
create or replace function tempus.delete_artificial_stop_road_section_f() returns trigger as $$
begin
    delete from tempus.road_node where id in (
        select node_from from tempus.road_section where id = OLD.road_section_id
        union all
        select node_to from tempus.road_section where id = OLD.road_section_id)
    ;
    delete from tempus.road_section where id = OLD.road_section_id;
    return OLD;
end;
$$ language plpgsql;

drop trigger if exists delete_artificial_stop_road_section on tempus_gtfs.stops;
create trigger delete_artificial_stop_road_section after delete on tempus_gtfs.stops
for each row
when (OLD.artificial_road_section)
execute procedure tempus.delete_artificial_stop_road_section_f();

-- link from transport_mode to gtfs feed
alter table tempus.transport_mode
add column gtfs_feed_id bigint references tempus_gtfs.feed_info(id) on delete cascade on update cascade;

-- shapes geometry
select AddGeometryColumn('tempus_gtfs', 'shapes', 'geom', 4326, 'LINESTRING', 3);

-- sections (edges between two stops)
create table tempus_gtfs.sections
(
	stop_from integer not null references tempus_gtfs.stops(id) on delete cascade on update cascade,
	stop_to integer not null references tempus_gtfs.stops(id) on delete cascade on update cascade,
	feed_id integer not null references tempus_gtfs.feed_info(id) on delete cascade on update cascade,
	primary key(stop_from, stop_to, feed_id)
);
select AddGeometryColumn('tempus_gtfs', 'sections', 'geom', 4326, 'LINESTRING', 3);

--
-- PostGIS geometry
--

--
-- Utilitary functions
--


-- find the closest road_section, then select the closest endpoint
create or replace function tempus.road_node_id_from_coordinates( float8, float8 ) returns bigint
as $$
with rs as (
select id, node_from, node_to from tempus.road_section
order by geom <-> st_setsrid(st_point($1, $2), 4326) limit 1
)
select case when st_distance( p1.geom, st_setsrid(st_point($1,$2), 4326)) < st_distance( p2.geom, st_setsrid(st_point($1,$2), 4326)) then p1.id else p2.id end
from rs, tempus.road_node as p1, tempus.road_node as p2
where
rs.node_from = p1.id and rs.node_to = p2.id
$$
language sql;

create or replace function tempus.road_node_id_from_coordinates_and_modes( float8, float8, int[] = array[1] ) returns bigint
as $$
with rs as (
select road_section.id, node_from, node_to from tempus.road_section, tempus.transport_mode
where
  transport_mode.id in (select unnest($3)) and
  (transport_mode.traffic_rules & traffic_rules_ft = transport_mode.traffic_rules
   or
  transport_mode.traffic_rules & traffic_rules_tf = transport_mode.traffic_rules)
order by geom <-> st_setsrid(st_point($1, $2), 4326) limit 1
)
select case when st_distance( p1.geom, st_setsrid(st_point($1,$2), 4326)) < st_distance( p2.geom, st_setsrid(st_point($1,$2), 4326)) then p1.id else p2.id end
from rs, tempus.road_node as p1, tempus.road_node as p2
where
rs.node_from = p1.id and rs.node_to = p2.id
$$
language sql;

DROP VIEW IF EXISTS tempus.view_forbidden_movements;
CREATE OR REPLACE VIEW tempus.view_forbidden_movements AS
 SELECT road_restriction.id,
    road_restriction.sections,
    st_union(road_section.geom) AS geom,
    max(road_restriction_time_penalty.traffic_rules) AS traffic_rules
   FROM tempus.road_section,
    tempus.road_restriction,
    tempus.road_restriction_time_penalty
  WHERE (road_section.id = ANY (road_restriction.sections)) AND road_restriction_time_penalty.restriction_id = road_restriction.id AND road_restriction_time_penalty.time_value = 'Infinity'::double precision
  GROUP BY road_restriction.id, road_restriction.sections;

DROP FUNCTION IF EXISTS tempus.array_search(anyelement, anyarray);

CREATE OR REPLACE FUNCTION tempus.array_search(needle anyelement, haystack anyarray)
  RETURNS integer AS
$BODY$
    SELECT i
      FROM generate_subscripts($2, 1) AS i
     WHERE $2[i] = $1
  ORDER BY i
$BODY$
  LANGUAGE sql STABLE
  COST 100;

-- for each pair of pt stops, departure, arrival_time and service_id of each available trip
create view tempus_gtfs.timetable as
select
  trips.feed_id,
  st1.id as origin_stop,
  st2.id as destination_stop,
  trips.id as trip_id,
  t1.departure_time / 60.0 as departure_time,
  t2.arrival_time / 60.0 as arrival_time,
  trips.service_id
from
  tempus_gtfs.stop_times t1
  join tempus_gtfs.stop_times t2 on (t1.trip_id = t2.trip_id) and (t1.feed_id = t2.feed_id) and (t2.stop_sequence = t1.stop_sequence + 1)
  join tempus_gtfs.trips on trips.trip_id = t1.trip_id and trips.feed_id = t1.feed_id
  join tempus_gtfs.stops st1 on st1.stop_id = t1.stop_id and st1.feed_id = t1.feed_id
  join tempus_gtfs.stops st2 on st2.stop_id = t2.stop_id and st2.feed_id = t2.feed_id
;

--
-- graph topology checks
--

create view tempus.chk_inconsistent_road_sections as
select
  rs.*
from
  tempus.road_section as rs
  left join tempus.road_node as rn1 on rs.node_from = rn1.id
  left join tempus.road_node as rn2 on rs.node_to = rn2.id
where
  rn1.id is null
  or
  rn2.id is null
;

-- cycles
create view tempus.chk_cycles as
select
  *
from
  tempus.road_section where node_from = node_to
;

create view tempus.chk_double_sections as
select
  rs1.id as rs1_id
  , rs2.id as rs2_id
from
  tempus.road_section rs1
  , tempus.road_section rs2
where
  rs1.id <> rs2.id
  and
  (
    ( rs1.node_from = rs2.node_to
      and
      rs1.node_to = rs2.node_from
      and
      rs1.traffic_rules_ft = rs2.traffic_rules_ft
      and
      rs1.traffic_rules_tf = rs2.traffic_rules_tf
    )
  or
    ( rs1.node_from = rs2.node_from
      and
      rs1.node_to = rs2.node_to
      and
      rs1.traffic_rules_ft = rs2.traffic_rules_tf
      and
      rs1.traffic_rules_tf = rs2.traffic_rules_ft
    )
  )
;

create view tempus.chk_isolated_stops as
select
  p.*
from
  tempus_gtfs.stops AS p
  left join tempus_gtfs.sections as s1 on p.id = s1.stop_from
  left join tempus_gtfs.sections as s2 on p.id = s2.stop_to
  left join tempus_gtfs.stop_times on p.stop_id = stop_times.stop_id
  left join tempus_gtfs.stops as pp on p.stop_id = pp.parent_station_id
where
  s1.stop_from is null
  and s2.stop_to is null
  and stop_times.stop_id is null
  and pp.parent_station_id is null
;

--
-- convenience view for data loading from core
--

-- pt stops with integer ids and network ids
create view tempus.load_stops as
select
  sections.feed_id as network_id
  , s.id
  , s.stop_name
  , s.location_type
  , p.id as parent_station_id
  , s.road_section_id
  , s.zone_id
  , s.abscissa_road_section
  , s.stop_lon as x
  , s.stop_lat as y
  , 0.0 as z  
from
  tempus_gtfs.sections
  join tempus_gtfs.stops s on s.id = stop_from
  left join tempus_gtfs.stops p on p.stop_id = s.parent_station_id
union
select
  sections.feed_id as network_id
  , s.id
  , s.stop_name
  , s.location_type
  , p.id as parent_station_id
  , s.road_section_id
  , s.zone_id
  , s.abscissa_road_section
  , s.stop_lon as x
  , s.stop_lat as y
  , 0.0 as z  
from
  tempus_gtfs.sections
  join tempus_gtfs.stops s on s.id = stop_to
  left join tempus_gtfs.stops p on p.stop_id = s.parent_station_id
;

-- subsets metadata

CREATE TABLE tempus.subset
(
	id serial PRIMARY KEY,
        schema_name text NOT NULL
);
-- polygon that has been used to define the subset
SELECT AddGeometryColumn('tempus', 'subset', 'geom', 4326, 'POLYGON', 2);

CREATE OR REPLACE FUNCTION tempus.create_subset(msubset text, polygon text) RETURNS void AS $$
DECLARE
    mdefinition text;
    -- declarations
BEGIN
  EXECUTE format('DROP SCHEMA IF EXISTS %s CASCADE', msubset);
  EXECUTE format('delete from geometry_columns where f_table_schema=''%s''', msubset);
  EXECUTE format('create schema %s', msubset);

  -- road nodes
  EXECUTE format('create table %s.road_node as select rn.* from tempus.road_node as rn where st_intersects( ''' || polygon || '''::geometry, rn.geom )', msubset);
  -- road sections
  EXECUTE format( 'create table %s.road_section as ' ||
	'select rs.* from tempus.road_section as rs ' ||
	'where ' ||
        'node_from in (select id from %1$s.road_node) ' ||
        'and node_to in (select id from %1$s.road_node)', msubset);
  -- pt stops
  EXECUTE format( 'create table %s.pt_stop as ' ||
                   'select pt.* from tempus.pt_stop as pt, %1$s.road_section as rs where road_section_id = rs.id', msubset );

  -- pt section
  EXECUTE format( 'create table %s.pt_section as ' ||
		'select pt.* from tempus.pt_section as pt ' ||
	        'where stop_from in (select id from %1$s.pt_stop) ' ||
                'and stop_to in (select id from %1$s.pt_stop)', msubset );

  -- pt stop time
  EXECUTE format( 'create table %s.pt_stop_time as ' ||
	'select st.* from tempus.pt_stop_time as st, %1$s.pt_stop as stop where stop_id = stop.id', msubset );

  -- poi
  EXECUTE format( 'create table %s.poi as ' ||
                  'select poi.* from tempus.poi as poi, %1$s.road_section as rs where road_section_id = rs.id', msubset );

  -- road_restriction
  EXECUTE format( 'create table %s.road_restriction as ' ||
         'select distinct rr.id, rr.sections from tempus.road_restriction as rr, %1$s.road_section as rs ' ||
         'where rs.id in (select unnest(sections) from tempus.road_restriction where id=rr.id)', msubset );

  EXECUTE format( 'create table %s.road_restriction_time_penalty as ' ||
         'select rrtp.* from tempus.road_restriction_time_penalty as rrtp, ' ||
         '%1$s.road_restriction as rr where restriction_id = rr.id', msubset );

  -- pt_trip
  EXECUTE format( 'create table %s.pt_trip as ' ||
                  'select * from tempus.pt_trip where id in (select distinct trip_id from %1$s.pt_stop_time union select trip_id from tempus.pt_frequency)', msubset);
  -- pt_frequency
  EXECUTE format( 'create table %s.pt_frequency as select * from tempus.pt_frequency', msubset );

  -- pt_route
  EXECUTE format( 'create table %s.pt_route as select * from tempus.pt_route where id in (select distinct route_id from %1$s.pt_trip)', msubset );

  -- pt_calendar
  EXECUTE format( 'create table %s.pt_calendar as select * from tempus.pt_calendar where service_id in (select distinct service_id from %1$s.pt_trip)', msubset );
  EXECUTE format( 'create table %s.pt_calendar_date as select * from tempus.pt_calendar_date where service_id in (select distinct service_id from %1$s.pt_trip)', msubset );

  -- pt_network
  EXECUTE format( 'create table %s.pt_network as select * from tempus.pt_network where id in (select distinct network_id from %1$s.pt_section)', msubset );
  -- pt_agency
  EXECUTE format( 'create table %s.pt_agency as select * from tempus.pt_agency where network_id in (select distinct id from %1$s.pt_network)', msubset );
  -- pt_zone
  EXECUTE format( 'create table %s.pt_zone as select * from tempus.pt_zone where network_id in (select distinct id from %1$s.pt_network)', msubset );
  -- pt_fare_rule
  EXECUTE format( 'create table %s.pt_fare_rule as select * from tempus.pt_fare_rule where route_id in (select id from %1$s.pt_route)', msubset );
  -- pt_fare_attribute
  EXECUTE format( 'create table %s.pt_fare_attribute as select * from tempus.pt_fare_attribute', msubset );

  -- transport_mode
  EXECUTE format( 'create table %s.transport_mode as select * from tempus.transport_mode', msubset );

  -- forbidden movements view
  SELECT replace(definition, 'tempus.', msubset || '.') into mdefinition from pg_views where schemaname='tempus' and viewname='view_forbidden_movements';
  EXECUTE 'create view ' || msubset || '.view_forbidden_movements as ' || mdefinition;

  -- update_pt_views
  select replace(prosrc, 'tempus.', msubset || '.') into mdefinition
     from pg_proc, pg_namespace where pg_proc.pronamespace = pg_namespace.oid and proname='update_pt_views' and nspname='tempus';
  EXECUTE 'create function ' || msubset || '.update_pt_views() returns void as $' || '$' || mdefinition || '$' || '$ language plpgsql';
  EXECUTE 'SELECT ' || msubset || '.update_pt_views()';

  DELETE FROM tempus.subset WHERE schema_name=msubset;
  INSERT INTO tempus.subset (schema_name, geom) VALUES (msubset, polygon::geometry);
END;
$$ LANGUAGE plpgsql;
